from __future__ import division

inf = float("inf")

from ._bracket import bracket
from ._brent import brent

_eps = 1.4902e-08

def minimize(f, x0=None, x1=None, a=-inf, b=+inf, gfactor=2,
             rtol=_eps, atol=_eps, maxiter=500):
    r"""Function minimization.

    Applies :func:`brent_search.bracket` and then :func:`brent_search.brent`
    to find the minimum.

    Args:
        f (callable): function of interest.
        x0 (:obj:`float`, optional): first point.
        x1 (:obj:`float`, optional): second point.
        a (:obj:`float`, optional): interval's lower limit. Defaults to ``-inf``.
        b (:obj:`float`, optional): interval's upper limit. Defaults to ``+inf``.
        gfactor (:obj:`float`, optional): growing factor.
        rtol (:obj:`float`, optional): relative tolerance. Defaults to ``1.4902e-08``.
        atol (:obj:`float`, optional): absolute tolerance. Defaults to ``1.4902e-08``.
        maxiter (:obj:`int`, optional): maximum number of iterations. Defaults to ``500``.

    Returns:
        A tuple containing the found solution (if any) in the first position,
        the function evaluated at that point, and the number of function
        evaluations.
    """

    def func(x):
        func.nfev += 1
        return f(x)
    func.nfev = 0


    r, _ = bracket(func, x0=x0, x1=x1, a=a, b=b, gfactor=gfactor, rtol=rtol,
                       atol=atol, maxiter=maxiter)

    x0, x1, x2, f0, f1, f2 = r[0], r[1], r[2], r[3], r[4], r[5]
    x0, f0 = brent(func, x0, x2, f0, f2, x1, f1, rtol, atol, maxiter)[0:2]
    return x0, f0, func.nfev



from __future__ import division

from numpy.testing import (assert_almost_equal, assert_array_less, assert_)

from brent_search import brent

def test_strictly_convex():
    def func(x, s):
        return (x - s)**2 - 0.8

    (x, fx, niters) = brent(lambda x: func(x, 0), -10, 10)
    assert_almost_equal(x, 0)
    assert_almost_equal(fx, -0.8)
    assert_array_less(niters, 7)

    (x, fx, niters) = brent(lambda x: func(x, 0), 0, 10)
    assert_almost_equal(x, 0)
    assert_array_less(niters, 42)

    (x, fx, niters) = brent(lambda x: func(x, 5), -1000, -100, rtol=1e-10)
    assert_almost_equal(x, -100)
    assert_almost_equal(fx, func(-100, 5), decimal=4)
    assert_array_less(niters, 51)

    (x, fx, niters) = brent(lambda x: func(x, 5), -100, -99, rtol=1e-10)
    assert_almost_equal(x, -99)
    assert_almost_equal(fx, func(-99, 5), decimal=4)

    (x, fx, niters) = brent(lambda x: func(x, 5), 99, 100, rtol=1e-10)
    assert_almost_equal(x, 99)
    assert_almost_equal(fx, func(99, 5), decimal=4)
    assert_array_less(niters, 36)

    (x, fx, niters) = brent(lambda x: func(x, 5), -1, -1/2)
    assert_almost_equal(x, -1/2)
    assert_almost_equal(fx, func(-1/2, 5), decimal=4)
    assert_array_less(niters, 35)

    (x, fx, niters) = brent(lambda x: func(x, 5), 6.5, 6.5)
    assert_almost_equal(x, 6.5)
    assert_almost_equal(fx, func(6.5, 5), decimal=4)
    assert_array_less(niters, 2)

    (x, fx, niters) = brent(lambda x: func(x, 5), 6, 7, rtol=1e-9)
    assert_almost_equal(x, 6)
    assert_almost_equal(fx, func(6, 5), decimal=4)
    assert_array_less(niters, 37)

def test_convex():
    def func(x, s):
        return abs(x - s) + 3

    (x, fx, niters) = brent(lambda x: func(x, 0), -10, +10)
    assert_almost_equal(x, 0)
    assert_almost_equal(fx, 3)
    assert_array_less(niters, 37)

    (x, fx, niters) = brent(lambda x: func(x, -9), -10, +10)
    assert_almost_equal(x, -9)
    assert_array_less(niters, 37)

def test_asymptotic():
    def func(x):
        return -3 + 1/x

    (x, _, niters) = brent(func, 1e-6, +10, rtol=1e-9)
    assert_almost_equal(x, 10)
    assert_array_less(niters, 42)

    (x, _, niters) = brent(func, -10, -1e-6)
    assert_almost_equal(x, -1e-06)
    assert_array_less(niters, 42)

def test_same_point():
    def func(x):
        return x**2
    (x, _, niters) = brent(func, 1.2, 1.2)
    assert_array_less(niters, 2)
    assert_almost_equal(x, 1.2)

def test_piecewise_convex():
    def func(x):
        if abs(x) > 2:
            return x**2
        if abs(x) > 1:
            return abs(x)
        return 1

    (x, fx, niters) = brent(func, -10, +10)
    assert_(-1-1e6 <= x <= 1+1e6)
    assert_almost_equal(fx, 1)
    assert_array_less(niters, 45)

    (x, fx, niters) = brent(func, 1/2, +1)
    assert_(-1-1e6 <= x <= 1+1e6)
    assert_almost_equal(fx, 1)
    assert_array_less(niters, 35)

    (x, fx, niters) = brent(func, -1, -1/2)
    assert_(-1-1e6 <= x <= 1+1e6)
    assert_almost_equal(fx, 1)
    assert_array_less(niters, 35)

    (x, fx, niters) = brent(func, -1/2, +1/2)
    assert_(-1-1e6 <= x <= 1+1e6)
    assert_almost_equal(fx, 1)
    assert_array_less(niters, 37)
